@@ -1,13 +1,16 @@
 import json
 from abc import ABC
+from copy import copy, deepcopy
 from enum import Enum
 from inspect import signature
+from logging import getLogger
 from typing import (
     TYPE_CHECKING,
     Any,
     Callable,
     Collection,
     Dict,
+    Iterable,
     Iterator,
     List,
     Mapping,
@@ -16,33 +19,35 @@ from typing import (
 )
 
 import opentelemetry.context as context_api
+from opentelemetry import trace as trace_api
+from opentelemetry.instrumentation.instrumentor import BaseInstrumentor  # type: ignore
+from opentelemetry.trace import StatusCode
+from opentelemetry.util.types import AttributeValue
+from wrapt import BoundFunctionWrapper, FunctionWrapper, wrap_object
+
+from openinference.instrumentation import (
+    OITracer,
+    TraceConfig,
+    get_attributes_from_context,
+    safe_json_dumps,
+)
 from openinference.instrumentation.dspy.package import _instruments
 from openinference.instrumentation.dspy.version import __version__
 from openinference.semconv.trace import (
     DocumentAttributes,
+    MessageAttributes,
     OpenInferenceMimeTypeValues,
     OpenInferenceSpanKindValues,
     SpanAttributes,
 )
-from opentelemetry import trace as trace_api
-from opentelemetry.instrumentation.instrumentor import BaseInstrumentor  # type: ignore
-from opentelemetry.util.types import AttributeValue
-from typing_extensions import TypeGuard
-from wrapt import wrap_function_wrapper
-
-try:
-    from google.generativeai.types import GenerateContentResponse  # type: ignore
-except ImportError:
-    GenerateContentResponse = None
 
 if TYPE_CHECKING:
-    from google.generativeai.types import (
-        GenerateContentResponse as GenerateContentResponseType,
-    )
+    from dspy import LM
+
+logger = getLogger(__name__)
+
 
 _DSPY_MODULE = "dspy"
-
-# DSPy used to be called DSP - some of the modules still fall under the old namespace
 _DSP_MODULE = "dsp"
 
 
@@ -57,61 +62,76 @@ class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
     def _instrument(self, **kwargs: Any) -> None:
         if not (tracer_provider := kwargs.get("tracer_provider")):
             tracer_provider = trace_api.get_tracer_provider()
-        tracer = trace_api.get_tracer(__name__, __version__, tracer_provider)
-
-        # Instrument LM (language model) calls
-        from dsp.modules.lm import LM
+        if not (config := kwargs.get("config")):
+            config = TraceConfig()
+        else:
+            assert isinstance(config, TraceConfig)
+        self._tracer = OITracer(
+            trace_api.get_tracer(__name__, __version__, tracer_provider),
+            config=config,
+        )
 
         from dspy import Predict
 
-        language_model_classes = LM.__subclasses__()
-        for lm in language_model_classes:
-            wrap_function_wrapper(
-                module=_DSP_MODULE,
-                name=lm.__name__ + ".basic_request",
-                wrapper=_LMBasicRequestWrapper(tracer),
-            )
+        wrap_object(
+            module="dspy",
+            name="LM.__call__",
+            factory=CopyableFunctionWrapper,
+            args=(_LMCallWrapper(self._tracer),),
+        )
 
         # Predict is a concrete (non-abstract) class that may be invoked
         # directly, but DSPy also has subclasses of Predict that override the
         # forward method. We instrument both the forward methods of the base
         # class and all subclasses.
-        wrap_function_wrapper(
+        wrap_object(
             module=_DSPY_MODULE,
             name="Predict.forward",
-            wrapper=_PredictForwardWrapper(tracer),
+            factory=CopyableFunctionWrapper,
+            args=(_PredictForwardWrapper(self._tracer),),
         )
 
         predict_subclasses = Predict.__subclasses__()
         for predict_subclass in predict_subclasses:
-            wrap_function_wrapper(
+            wrap_object(
                 module=_DSPY_MODULE,
                 name=predict_subclass.__name__ + ".forward",
-                wrapper=_PredictForwardWrapper(tracer),
+                factory=CopyableFunctionWrapper,
+                args=(_PredictForwardWrapper(self._tracer),),
             )
 
-        wrap_function_wrapper(
+        wrap_object(
             module=_DSPY_MODULE,
             name="Retrieve.forward",
-            wrapper=_RetrieverForwardWrapper(tracer),
+            factory=CopyableFunctionWrapper,
+            args=(_RetrieverForwardWrapper(self._tracer),),
         )
 
-        wrap_function_wrapper(
+        wrap_object(
             module=_DSPY_MODULE,
             # At this time, dspy.Module does not have an abstract forward
             # method, but assumes that user-defined subclasses implement the
             # forward method and invokes that method using __call__.
             name="Module.__call__",
-            wrapper=_ModuleForwardWrapper(tracer),
+            factory=CopyableFunctionWrapper,
+            args=(_ModuleForwardWrapper(self._tracer),),
         )
 
         # At this time, there is no common parent class for retriever models as
         # there is for language models. We instrument the retriever models on a
         # case-by-case basis.
-        wrap_function_wrapper(
+        wrap_object(
             module=_DSP_MODULE,
             name="ColBERTv2.__call__",
-            wrapper=_RetrieverModelCallWrapper(tracer),
+            factory=CopyableFunctionWrapper,
+            args=(_RetrieverModelCallWrapper(self._tracer),),
+        )
+
+        wrap_object(
+            module=_DSPY_MODULE,
+            name="Adapter.__call__",
+            factory=CopyableFunctionWrapper,
+            args=(_AdapterCallWrapper(self._tracer),),
         )
 
     def _uninstrument(self, **kwargs: Any) -> None:
@@ -129,6 +149,49 @@ class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
             Predict.forward = Predict.forward.__wrapped__
 
 
+class CopyableBoundFunctionWrapper(BoundFunctionWrapper):  # type: ignore
+    """
+    A bound function wrapper that can be copied and deep-copied. When used to
+    wrap a class method, this allows the entire class to be copied and
+    deep-copied.
+
+    For reference, see
+    https://github.com/GrahamDumpleton/wrapt/issues/86#issuecomment-426161271
+    and
+    https://wrapt.readthedocs.io/en/master/wrappers.html#custom-function-wrappers
+    """
+
+    def __copy__(self) -> "CopyableBoundFunctionWrapper":
+        return CopyableBoundFunctionWrapper(
+            copy(self.__wrapped__), self._self_instance, self._self_wrapper
+        )
+
+    def __deepcopy__(self, memo: Dict[Any, Any]) -> "CopyableBoundFunctionWrapper":
+        return CopyableBoundFunctionWrapper(
+            deepcopy(self.__wrapped__, memo), self._self_instance, self._self_wrapper
+        )
+
+
+class CopyableFunctionWrapper(FunctionWrapper):  # type: ignore
+    """
+    A function wrapper that can be copied and deep-copied. When used to wrap a
+    class method, this allows the entire class to be copied and deep-copied.
+
+    For reference, see
+    https://github.com/GrahamDumpleton/wrapt/issues/86#issuecomment-426161271
+    and
+    https://wrapt.readthedocs.io/en/master/wrappers.html#custom-function-wrappers
+    """
+
+    __bound_function_wrapper__ = CopyableBoundFunctionWrapper
+
+    def __copy__(self) -> "CopyableFunctionWrapper":
+        return CopyableFunctionWrapper(copy(self.__wrapped__), self._self_wrapper)
+
+    def __deepcopy__(self, memo: Dict[Any, Any]) -> "CopyableFunctionWrapper":
+        return CopyableFunctionWrapper(deepcopy(self.__wrapped__, memo), self._self_wrapper)
+
+
 class _WithTracer(ABC):
     """
     Base class for wrappers that need a tracer. Acts as a trait for the wrappers
@@ -139,57 +202,49 @@ class _WithTracer(ABC):
         self._tracer = tracer
 
 
-class _LMBasicRequestWrapper(_WithTracer):
+class _LMCallWrapper(_WithTracer):
     """
-    Wrapper for DSP LM.basic_request
-    Captures all calls to language models (lm)
+    Wrapper for __call__ method on dspy.LM
     """
 
     def __call__(
         self,
         wrapped: Callable[..., Any],
-        instance: Any,
+        instance: "LM",
         args: Tuple[type, Any],
         kwargs: Mapping[str, Any],
     ) -> Any:
         if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
             return wrapped(*args, **kwargs)
-        prompt = args[0]
-        invocation_parameters = {**instance.kwargs, **kwargs}
-        span_name = instance.__class__.__name__ + ".request"
+        arguments = _bind_arguments(wrapped, *args, **kwargs)
+        span_name = instance.__class__.__name__ + ".__call__"
         with self._tracer.start_as_current_span(
             span_name,
             attributes=dict(
                 _flatten(
                     {
-                        OPENINFERENCE_SPAN_KIND: LLM.value,
-                        LLM_MODEL_NAME: instance.kwargs.get("model"),
-                        LLM_INVOCATION_PARAMETERS: json.dumps(invocation_parameters),
-                        INPUT_VALUE: str(prompt),
-                        INPUT_MIME_TYPE: OpenInferenceMimeTypeValues.TEXT.value,
+                        OPENINFERENCE_SPAN_KIND: LLM,
+                        **dict(_input_value_and_mime_type(arguments)),
+                        **dict(_llm_model_name(instance)),
+                        **dict(_llm_invocation_parameters(instance, arguments)),
+                        **dict(_llm_input_messages(arguments)),
+                        **dict(get_attributes_from_context()),
                     }
                 )
             ),
         ) as span:
-            try:
-                response = wrapped(*args, **kwargs)
-            except Exception as exception:
-                span.set_status(trace_api.Status(trace_api.StatusCode.ERROR, str(exception)))
-                span.record_exception(exception)
-                raise
-            # TODO: parse usage. Need to decide if this
-            # instrumentation should be used in conjunction with model instrumentation
+            response = wrapped(*args, **kwargs)
+            span.set_status(StatusCode.OK)
             span.set_attributes(
                 dict(
                     _flatten(
                         {
-                            OUTPUT_VALUE: _jsonify_output(response),
-                            OUTPUT_MIME_TYPE: OpenInferenceMimeTypeValues.JSON.value,
+                            **dict(_output_value_and_mime_type(response)),
+                            **dict(_llm_output_messages(response)),
                         }
                     )
                 )
             )
-            span.set_status(trace_api.StatusCode.OK)
         return response
 
 
@@ -239,36 +294,32 @@ class _PredictForwardWrapper(_WithTracer):
             attributes=dict(
                 _flatten(
                     {
-                        OPENINFERENCE_SPAN_KIND: CHAIN.value,
+                        OPENINFERENCE_SPAN_KIND: CHAIN,
                         INPUT_VALUE: _get_input_value(
                             wrapped,
                             *args,
                             **kwargs,
                         ),
-                        INPUT_MIME_TYPE: OpenInferenceMimeTypeValues.JSON.value,
+                        INPUT_MIME_TYPE: JSON,
                     }
                 )
             ),
         ) as span:
-            try:
-                prediction = wrapped(*args, **kwargs)
-            except Exception as exception:
-                span.set_status(trace_api.Status(trace_api.StatusCode.ERROR, str(exception)))
-                span.record_exception(exception)
-                raise
+            span.set_attributes(dict(get_attributes_from_context()))
+            prediction = wrapped(*args, **kwargs)
             span.set_attributes(
                 dict(
                     _flatten(
                         {
-                            OUTPUT_VALUE: json.dumps(
+                            OUTPUT_VALUE: safe_json_dumps(
                                 self._prediction_to_output_dict(prediction, signature)
                             ),
-                            OUTPUT_MIME_TYPE: OpenInferenceMimeTypeValues.JSON.value,
+                            OUTPUT_MIME_TYPE: JSON,
                         }
                     )
                 )
             )
-            span.set_status(trace_api.StatusCode.OK)
+            span.set_status(StatusCode.OK)
         return prediction
 
     def _prediction_to_output_dict(self, prediction: Any, signature: Any) -> Dict[str, Any]:
@@ -305,7 +356,7 @@ class _ModuleForwardWrapper(_WithTracer):
             attributes=dict(
                 _flatten(
                     {
-                        OPENINFERENCE_SPAN_KIND: CHAIN.value,
+                        OPENINFERENCE_SPAN_KIND: CHAIN,
                         # At this time, dspy.Module does not have an abstract forward
                         # method, but assumes that user-defined subclasses implement the
                         # forward method.
@@ -314,28 +365,24 @@ class _ModuleForwardWrapper(_WithTracer):
                             if (forward_method := getattr(instance.__class__, "forward", None))
                             else {}
                         ),
-                        INPUT_MIME_TYPE: OpenInferenceMimeTypeValues.JSON.value,
+                        INPUT_MIME_TYPE: JSON,
                     }
                 )
             ),
         ) as span:
-            try:
-                prediction = wrapped(*args, **kwargs)
-            except Exception as exception:
-                span.set_status(trace_api.Status(trace_api.StatusCode.ERROR, str(exception)))
-                span.record_exception(exception)
-                raise
+            span.set_attributes(dict(get_attributes_from_context()))
+            prediction = wrapped(*args, **kwargs)
             span.set_attributes(
                 dict(
                     _flatten(
                         {
-                            OUTPUT_VALUE: json.dumps(prediction, cls=DSPyJSONEncoder),
-                            OUTPUT_MIME_TYPE: OpenInferenceMimeTypeValues.JSON.value,
+                            OUTPUT_VALUE: safe_json_dumps(prediction, cls=DSPyJSONEncoder),
+                            OUTPUT_MIME_TYPE: JSON,
                         }
                     )
                 )
             )
-            span.set_status(trace_api.StatusCode.OK)
+            span.set_status(StatusCode.OK)
         return prediction
 
 
@@ -365,17 +412,13 @@ class _RetrieverForwardWrapper(_WithTracer):
                     {
                         OPENINFERENCE_SPAN_KIND: RETRIEVER.value,
                         INPUT_VALUE: _get_input_value(wrapped, *args, **kwargs),
-                        INPUT_MIME_TYPE: OpenInferenceMimeTypeValues.JSON.value,
+                        INPUT_MIME_TYPE: JSON,
                     }
                 )
             ),
         ) as span:
-            try:
-                prediction = wrapped(*args, **kwargs)
-            except Exception as exception:
-                span.set_status(trace_api.Status(trace_api.StatusCode.ERROR, str(exception)))
-                span.record_exception(exception)
-                raise
+            span.set_attributes(dict(get_attributes_from_context()))
+            prediction = wrapped(*args, **kwargs)
             span.set_attributes(
                 dict(
                     _flatten(
@@ -390,7 +433,7 @@ class _RetrieverForwardWrapper(_WithTracer):
                     )
                 )
             )
-            span.set_status(trace_api.StatusCode.OK)
+            span.set_status(StatusCode.OK)
         return prediction
 
 
@@ -417,17 +460,13 @@ class _RetrieverModelCallWrapper(_WithTracer):
                     {
                         OPENINFERENCE_SPAN_KIND: RETRIEVER.value,
                         INPUT_VALUE: (_get_input_value(wrapped, *args, **kwargs)),
-                        INPUT_MIME_TYPE: OpenInferenceMimeTypeValues.JSON.value,
+                        INPUT_MIME_TYPE: JSON,
                     }
                 )
             ),
         ) as span:
-            try:
-                retrieved_documents = wrapped(*args, **kwargs)
-            except Exception as exception:
-                span.set_status(trace_api.Status(trace_api.StatusCode.ERROR, str(exception)))
-                span.record_exception(exception)
-                raise
+            span.set_attributes(dict(get_attributes_from_context()))
+            retrieved_documents = wrapped(*args, **kwargs)
             span.set_attributes(
                 dict(
                     _flatten(
@@ -444,10 +483,48 @@ class _RetrieverModelCallWrapper(_WithTracer):
                     )
                 )
             )
-            span.set_status(trace_api.StatusCode.OK)
+            span.set_status(StatusCode.OK)
         return retrieved_documents
 
 
+class _AdapterCallWrapper(_WithTracer):
+    def __call__(
+        self,
+        wrapped: Callable[..., Any],
+        instance: Any,
+        args: Tuple[type, Any],
+        kwargs: Mapping[str, Any],
+    ) -> Any:
+        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
+            return wrapped(*args, **kwargs)
+        arguments = _bind_arguments(wrapped, *args, **kwargs)
+        span_name = instance.__class__.__name__ + ".__call__"
+        with self._tracer.start_as_current_span(
+            span_name,
+            attributes=dict(
+                _flatten(
+                    {
+                        OPENINFERENCE_SPAN_KIND: CHAIN,
+                        **dict(_input_value_and_mime_type(arguments)),
+                        **dict(get_attributes_from_context()),
+                    }
+                )
+            ),
+        ) as span:
+            response = wrapped(*args, **kwargs)
+            span.set_status(StatusCode.OK)
+            span.set_attributes(
+                dict(
+                    _flatten(
+                        {
+                            **dict(_output_value_and_mime_type(response)),
+                        }
+                    )
+                )
+            )
+        return response
+
+
 class DSPyJSONEncoder(json.JSONEncoder):
     """
     Provides support for non-JSON-serializable objects in DSPy.
@@ -475,6 +552,20 @@ class DSPyJSONEncoder(json.JSONEncoder):
             return repr(o)
 
 
+class SafeJSONEncoder(json.JSONEncoder):
+    """
+    Safely encodes non-JSON-serializable objects.
+    """
+
+    def default(self, o: Any) -> Any:
+        try:
+            return super().default(o)
+        except TypeError:
+            if hasattr(o, "dict") and callable(o.dict):  # pydantic v1 models, e.g., from Cohere
+                return o.dict()
+            return repr(o)
+
+
 def _get_input_value(method: Callable[..., Any], *args: Any, **kwargs: Any) -> str:
     """
     Parses a method call's inputs into a JSON string. Ensures a consistent
@@ -497,7 +588,7 @@ def _get_input_value(method: Callable[..., Any], *args: Any, **kwargs: Any) -> s
         *args,
         **kwargs,
     )
-    return json.dumps(
+    return safe_json_dumps(
         {
             **{
                 argument_name: argument_value
@@ -554,42 +645,70 @@ def _flatten(mapping: Mapping[str, Any]) -> Iterator[Tuple[str, AttributeValue]]
             yield key, value
 
 
-def _jsonify_output(response: Any) -> str:
-    """
-    Converts output to JSON string.
-    """
-    if _is_google_response(response):
-        return json.dumps(_parse_google_response(response))
-    return json.dumps(response)
+def _input_value_and_mime_type(arguments: Mapping[str, Any]) -> Iterator[Tuple[str, Any]]:
+    yield INPUT_MIME_TYPE, JSON
+    yield INPUT_VALUE, safe_json_dumps(arguments)
 
 
-def _is_google_response(response: Any) -> TypeGuard["GenerateContentResponseType"]:
-    """
-    Checks whether a candidate response is an instance of
-    GenerateContentResponse returned by the Google generative AI SDK.
-    """
-
-    return GenerateContentResponse is not None and isinstance(response, GenerateContentResponse)
+def _output_value_and_mime_type(response: Any) -> Iterator[Tuple[str, Any]]:
+    yield OUTPUT_VALUE, safe_json_dumps(response)
+    yield OUTPUT_MIME_TYPE, JSON
 
 
-def _parse_google_response(response: "GenerateContentResponseType") -> Dict[str, Any]:
-    """
-    Parses a response from the Google generative AI SDK into a dictionary.
-    """
-
-    return {
-        "text": response.text,
-    }
+def _llm_model_name(lm: "LM") -> Iterator[Tuple[str, Any]]:
+    if (model_name := getattr(lm, "model_name", None)) is not None:
+        yield LLM_MODEL_NAME, model_name
 
 
+def _llm_input_messages(arguments: Mapping[str, Any]) -> Iterator[Tuple[str, Any]]:
+    if isinstance(prompt := arguments.get("prompt"), str):
+        yield f"{LLM_INPUT_MESSAGES}.0.{MESSAGE_ROLE}", "user"
+        yield f"{LLM_INPUT_MESSAGES}.0.{MESSAGE_CONTENT}", prompt
+    elif isinstance(messages := arguments.get("messages"), list):
+        for i, message in enumerate(messages):
+            if not isinstance(message, dict):
+                continue
+            if (role := message.get("role", None)) is not None:
+                yield f"{LLM_INPUT_MESSAGES}.{i}.{MESSAGE_ROLE}", role
+            if (content := message.get("content", None)) is not None:
+                yield f"{LLM_INPUT_MESSAGES}.{i}.{MESSAGE_CONTENT}", content
+
+
+def _llm_output_messages(response: Any) -> Iterator[Tuple[str, Any]]:
+    if isinstance(response, Iterable):
+        for i, message in enumerate(response):
+            if isinstance(message, str):
+                yield f"{LLM_OUTPUT_MESSAGES}.{i}.{MESSAGE_ROLE}", "assistant"
+                yield f"{LLM_OUTPUT_MESSAGES}.{i}.{MESSAGE_CONTENT}", message
+
+
+def _llm_invocation_parameters(lm: "LM", arguments: Mapping[str, Any]) -> Iterator[Tuple[str, Any]]:
+    lm_kwargs = _ if isinstance(_ := getattr(lm, "kwargs", {}), dict) else {}
+    kwargs = _ if isinstance(_ := arguments.get("kwargs"), dict) else {}
+    yield LLM_INVOCATION_PARAMETERS, safe_json_dumps(lm_kwargs | kwargs)
+
+
+def _bind_arguments(method: Callable[..., Any], *args: Any, **kwargs: Any) -> Dict[str, Any]:
+    method_signature = signature(method)
+    bound_args = method_signature.bind(*args, **kwargs)
+    bound_args.apply_defaults()
+    return bound_args.arguments
+
+
+JSON = OpenInferenceMimeTypeValues.JSON.value
+TEXT = OpenInferenceMimeTypeValues.TEXT.value
+LLM = OpenInferenceSpanKindValues.LLM
 OPENINFERENCE_SPAN_KIND = SpanAttributes.OPENINFERENCE_SPAN_KIND
 RETRIEVER = OpenInferenceSpanKindValues.RETRIEVER
-CHAIN = OpenInferenceSpanKindValues.CHAIN
-LLM = OpenInferenceSpanKindValues.LLM
+CHAIN = OpenInferenceSpanKindValues.CHAIN.value
 INPUT_VALUE = SpanAttributes.INPUT_VALUE
 INPUT_MIME_TYPE = SpanAttributes.INPUT_MIME_TYPE
 OUTPUT_VALUE = SpanAttributes.OUTPUT_VALUE
 OUTPUT_MIME_TYPE = SpanAttributes.OUTPUT_MIME_TYPE
 LLM_INVOCATION_PARAMETERS = SpanAttributes.LLM_INVOCATION_PARAMETERS
+LLM_INPUT_MESSAGES = SpanAttributes.LLM_INPUT_MESSAGES
+LLM_OUTPUT_MESSAGES = SpanAttributes.LLM_OUTPUT_MESSAGES
 LLM_MODEL_NAME = SpanAttributes.LLM_MODEL_NAME
 RETRIEVAL_DOCUMENTS = SpanAttributes.RETRIEVAL_DOCUMENTS
+MESSAGE_CONTENT = MessageAttributes.MESSAGE_CONTENT
+MESSAGE_ROLE = MessageAttributes.MESSAGE_ROLE
